/*
 * ui.c
 *
 * User interface
 */

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <commdlg.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include "resource.h"
#include "bsp.h"

#define WALL_HEIGHT	70


/*****************************************************************************
 Data
*****************************************************************************/

static LINE		*line_list = NULL;	// last element inserted into beginning

static HMENU	menu;				// our menu


/*****************************************************************************
 Utility Functions 
*****************************************************************************/

static void Error(char *msg)
{
	MessageBox(NULL, msg, "Error", MB_ICONERROR);
}

static void DeleteAll()	// frees all lines
{
	LINE	*l;

	while (line_list != NULL)
	{
		l = line_list->next;
		free(line_list);
		line_list = l;
	}
}

static void AddLine(int x1, int y1, int x2, int y2)
{
	LINE	*n;

	if ((n = malloc(sizeof(LINE))) == NULL)
	{
		Error("Out of memory! You may be able to save your work as a line list, but not as a BSP.");
		return;
	}

	n->x1 = x1;
	n->y1 = y1;
	n->x2 = x2;
	n->y2 = y2;
	n->next = line_list;
	line_list = n;	// insert at beginning of list

	EnableMenuItem(menu, IDM_EDIT_UNDO, MF_ENABLED);
	EnableMenuItem(menu, IDM_EDIT_CLEARALL, MF_ENABLED);
}

static BOOL SaveFile(HWND hwnd, OPENFILENAME *ofn, char *file, char *file_title, char *filter, char *defext, char *defname)
{     
 	strcpy(file, defname);
	
	memset(ofn, 0, sizeof(OPENFILENAME));
    ofn->lStructSize = sizeof(OPENFILENAME);
    ofn->hwndOwner = hwnd;
	ofn->hInstance = NULL;
    ofn->lpstrFilter = filter;
    ofn->nFilterIndex = 0;
	ofn->lpstrCustomFilter = NULL;
	ofn->nMaxCustFilter = 0;
	ofn->nFilterIndex = 0;
    ofn->lpstrFile = file;
    ofn->nMaxFile = MAX_PATH;
    ofn->lpstrFileTitle = file_title;
    ofn->nMaxFileTitle = MAX_PATH;
    ofn->lpstrInitialDir = NULL;
	ofn->lpstrDefExt = defext;
    ofn->Flags = OFN_OVERWRITEPROMPT;

	return GetSaveFileName(ofn);
}

static void SaveAsList(HWND hwnd)
{
    char			file[MAX_PATH], file_title[MAX_PATH];
	OPENFILENAME	ofn;	
	FILE			*fp;
	LINE			*l;

 	if (SaveFile(hwnd, &ofn, file, file_title, "Line List (*.lin)\0*.lin\0\0", "lin", "lines.lin"))	// open the file for writing
	{
		if ((fp = fopen(ofn.lpstrFile, "w")) == NULL)
		{			
			Error("Failed to open output line list file.");
            return;
	    }

		l = line_list;
		while (l != NULL)
		{
			fprintf(fp, "%d,%d-%d,%d\n", l->x1, l->y1, l->x2, l->y2);
			fprintf(fp, "%d,%d-%d,%d\n", l->x2, l->y2, l->x1, l->y1);
			l = l->next;
		}
		fprintf(fp, "0,0-0,0");
    	
        fclose(fp);
    }
}

static void SaveAsBSP(HWND hwnd)
{
    char			file[MAX_PATH], file_title[MAX_PATH];
	OPENFILENAME	ofn;	
	FILE			*fp;

 	if (SaveFile(hwnd, &ofn, file, file_title, "BSP Tree (*.bsp)\0*.bsp\0\0", "bsp", "world.bsp"))	// open the file for writing
	{
		if ((fp = fopen(ofn.lpstrFile, "w")) == NULL)
		{			
			Error("Failed to open output BSP tree file.");
            return;
	    }

		if (MakeBSP(fp, line_list))
			Error("Failed to write BSP tree!");

        fclose(fp);
    }
}

static void SaveAsPolyList(HWND hwnd)
{
    char			file[MAX_PATH], file_title[MAX_PATH];
	OPENFILENAME	ofn;	
	FILE			*fp;
	LINE			*l;

 	if (SaveFile(hwnd, &ofn, file, file_title, "Polygon List (*.pol)\0*.pol\0\0", "lin", "poly.pol"))	// open the file for writing
	{
		if ((fp = fopen(ofn.lpstrFile, "w")) == NULL)
		{			
			Error("Failed to open output line list file.");
            return;
	    }

		l = line_list;
		while (l != NULL)
		{
			fprintf(fp, "4\n");
			fprintf(fp, "%d,%d,%d\n", l->x1, -WALL_HEIGHT, l->y1);
			fprintf(fp, "%d,%d,%d\n", l->x2, -WALL_HEIGHT, l->y2);
			fprintf(fp, "%d,%d,%d\n", l->x2, WALL_HEIGHT, l->y2);
			fprintf(fp, "%d,%d,%d\n", l->x1, WALL_HEIGHT, l->y1);

			fprintf(fp, "4\n");
			fprintf(fp, "%d,%d,%d\n", l->x1, -WALL_HEIGHT, l->y1);
			fprintf(fp, "%d,%d,%d\n", l->x1, WALL_HEIGHT, l->y1);
			fprintf(fp, "%d,%d,%d\n", l->x2, WALL_HEIGHT, l->y2);
			fprintf(fp, "%d,%d,%d\n", l->x2, -WALL_HEIGHT, l->y2);
			
			l = l->next;
		}
		fprintf(fp, "0");
    	
        fclose(fp);
    }
}


/*****************************************************************************
 Windows Main Code

 WinMain(), the message pump, the message handler, and what-not.
*****************************************************************************/

/*
 * Windows message callback
 */

LRESULT CALLBACK WndProc(HWND hwnd, UINT msg, WPARAM wparam, LPARAM lparam)
{						
	HDC					hdc;
	PAINTSTRUCT			ps;
	LINE				*l;
	static HPEN			white_pen, black_pen;
	static int			start_x, start_y;		// start position of current line
	static int			old_end_x, old_end_y;	// old end position
	
			
	switch(msg)
	{
	case WM_CREATE:
		white_pen = GetStockObject(WHITE_PEN);
		black_pen = GetStockObject(BLACK_PEN);
		return 0;
	case WM_PAINT:
		hdc = BeginPaint(hwnd, &ps);

		/*
		 * Draw all lines
		 */

		l = line_list;
		while (l != NULL)
		{
			MoveToEx(hdc, l->x1, l->y1, NULL);
			LineTo(hdc, l->x2, l->y2);
			l = l->next;
		}

		EndPaint(hwnd, &ps);
		
		return 0;
	case WM_DESTROY:
		PostQuitMessage(0);
		return 0;

	case WM_LBUTTONDOWN:		
		start_x = old_end_x = LOWORD(lparam);
		start_y = old_end_y = HIWORD(lparam);
		return 0;
	case WM_LBUTTONUP:			
		hdc = BeginPaint(hwnd, &ps);
		
		MoveToEx(hdc, start_x, start_y, NULL);
		LineTo(hdc, LOWORD(lparam), HIWORD(lparam));

		EndPaint(hwnd, &ps);
		
		AddLine(start_x, start_y, LOWORD(lparam), HIWORD(lparam));
		InvalidateRect(hwnd, NULL, TRUE);
		return 0;

	case WM_MOUSEMOVE:
		if (!(wparam & MK_LBUTTON))
			return 0;

		InvalidateRect(hwnd, NULL, FALSE);
		hdc = BeginPaint(hwnd, &ps);

		SelectObject(hdc, white_pen);
		MoveToEx(hdc, start_x, start_y, NULL);
		LineTo(hdc, old_end_x, old_end_y);

		old_end_x = LOWORD(lparam);				// the new "old" coordinates
		old_end_y = HIWORD(lparam);

		SelectObject(hdc, black_pen);
		MoveToEx(hdc, start_x, start_y, NULL);
		LineTo(hdc, old_end_x, old_end_y);

		EndPaint(hwnd, &ps);
		InvalidateRect(hwnd, NULL, FALSE);
		return 0;

	case WM_COMMAND:
		switch (LOWORD(wparam))
		{
		case IDM_FILE_SAVEASLIST:
			SaveAsList(hwnd);
			return 0;
		case IDM_FILE_SAVEASBSP:
			SaveAsBSP(hwnd);
			return 0;
		case IDM_FILE_SAVEASPOLYLIST:
			SaveAsPolyList(hwnd);
			return 0;
		case IDM_FILE_EXIT:			
			SendMessage(hwnd, WM_CLOSE, 0, 0);
			return 0;
		case IDM_EDIT_UNDO:
			if (line_list != NULL)
			{
				l = line_list->next;
				free(line_list);
				line_list = l;

				if (line_list == NULL)
				{
					EnableMenuItem(menu, IDM_EDIT_UNDO, MF_GRAYED);
					EnableMenuItem(menu, IDM_EDIT_CLEARALL, MF_GRAYED);
				}

				InvalidateRect(hwnd, NULL, TRUE);
			}
			return 0;
		case IDM_EDIT_CLEARALL:
			DeleteAll();
			EnableMenuItem(menu, IDM_EDIT_UNDO, MF_GRAYED);
			EnableMenuItem(menu, IDM_EDIT_CLEARALL, MF_GRAYED);
			InvalidateRect(hwnd, NULL, TRUE);
			return 0;
		}
		break;
	}
	return DefWindowProc(hwnd, msg, wparam, lparam);
}

/*
 * int WINAPI WinMain(HINSTANCE hinstance, HINSTANCE hprevinstance, 
 *                    LPSTR args, int show_cmnd);
 *
 * Program execution begins here.
 */

int WINAPI WinMain(HINSTANCE hinstance, HINSTANCE hprevinstance, LPSTR args, int show_cmnd)
{
	static char	app_name[] = "bspmake";
	HWND		hwnd;	
	WNDCLASS	wndclass;
	MSG			msg;

	/* 
	 * Set the application options 
	 */

	wndclass.style			= CS_HREDRAW | CS_VREDRAW;
	wndclass.lpfnWndProc	= WndProc;
	wndclass.cbClsExtra		= 0;
	wndclass.cbWndExtra		= 0;
	wndclass.hInstance		= hinstance;
	wndclass.hIcon			= LoadIcon(NULL, IDI_APPLICATION);
	wndclass.hCursor		= LoadCursor(NULL, IDC_ARROW);
	wndclass.hbrBackground	= (HBRUSH) GetStockObject(WHITE_BRUSH);
	wndclass.lpszMenuName	= app_name;
	wndclass.lpszClassName	= app_name;

	/* 
	 * Register the window
	 */

	if (!RegisterClass(&wndclass))
	{
		MessageBox(NULL, "Failed to register window class!","Error",0); 
		return 0;
	}

	/*
	 * Create the window
	 */

	hwnd = CreateWindow(app_name,		// window class name
						"bspmake",		// window caption
						WS_OVERLAPPEDWINDOW,
						CW_USEDEFAULT,
						CW_USEDEFAULT,
						500,
						500,
						NULL,
						NULL,
						hinstance,
						NULL);

	
	/*
	 * Initialization
	 */

	menu = GetMenu(hwnd);

	/* 
	 * Display the window
	 */

	ShowWindow(hwnd, show_cmnd);
	UpdateWindow(hwnd);

	
	/* 
	 * Ye Olde Message Pumpe
	 */
	
	while (GetMessage(&msg, NULL, 0, 0))
	{
		TranslateMessage(&msg);
		DispatchMessage(&msg);
	}

	/*
	 * Clean up and quit
	 */

	DeleteAll();

	return msg.wParam;
}