#define FOV 90.0

/*
 * r.c
 *
 * BSP renderer
 */

#include <conio.h>
#include <dpmi.h>
#include <go32.h>
#include <math.h>
#include <pc.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/farptr.h>
#include "types.h"
#include "bsp.h"
#include "keyb.h"
#include "poly.h"
#include "scrshot.h"
#include "vesalib.h"
#include "vector.h"


#define WALLHEIGHT  70

typedef struct
{
    float   x, y, z;
} VERTEX3D;


/*****************************************************************************
 Data
*****************************************************************************/

/*
 * Node List
 */

static NODE *node_list = NULL;

/*
 * BSP Tree
 */

static BSP  *tree;

/*
 * Projection and FOV Stuff
 */

static float    dphys;
static float    fov_y;

/*
 * View Position and Angle
 */

static float    x = 0, z = -1000;
static float    v_a = 0.0;


/*
 * Wall Color
 */

static int      color;


/*****************************************************************************
 Video Data
*****************************************************************************/

/*
 * Screen Dimensions
 */

static INT32    xres;
static INT32    yres;

/*
 * Frame Buffer
 */

static UINT8            *vidbuf;
static VESALIB_LFBINFO  lfb;


/*****************************************************************************
 Clipping
*****************************************************************************/

#define VERT2VEC(v, p_vert) { v.x = p_vert->x; v.y = p_vert->y; v.z = p_vert->z; }

static INT32 ClipPolyToPlane(VERTEX3D *outlist, VERTEX3D *inlist, PLANE *plane, INT32 num_vertices)
{  
    VECTOR3D    v;
    INT32       i, current_in, next_in, next_i, out_vertices;
    float       current_dot, next_dot, scale;
    VERTEX3D    *out, *in;

    out = outlist;  // for easy access
    in = inlist;

    // find distance of current vertex

    VERT2VEC(v, in);
    current_dot = Vec3D_RawDotProduct(&(plane->unit_normal), &v);
    current_in = (current_dot <= plane->dist);  // 1 if inside plane

    for (i = out_vertices = 0; i < num_vertices; i++)
    {
        if (current_in) // if the current vertex is inside, keep it
        {
            *out++ = *in;
            out_vertices++;
        }

        next_i = i + 1;             // next vertex in polygon
        if (next_i >= num_vertices)
            next_i = 0;

        VERT2VEC(v, (&inlist[next_i]));
        next_dot = Vec3D_RawDotProduct(&(plane->unit_normal), &v);
        next_in = (next_dot <= plane->dist);    // 1 if inside plane

        if (next_in != current_in)  // there is an intersection somewhere!
        {
            scale = (plane->dist - current_dot) / (next_dot - current_dot);

            out->x = in->x + (inlist[next_i].x - in->x) * scale;
            out->y = in->y + (inlist[next_i].y - in->y) * scale;
            out->z = in->z + (inlist[next_i].z - in->z) * scale;

            out++;
            out_vertices++;
        }

        current_dot = next_dot;
        current_in = next_in;
        in++;        
    }

    return out_vertices;
}

static INT32 Clip(VERTEX3D *v)
{
    VERTEX3D    v_2[4*4];
    PLANE       frustum[4];
    VECTOR3D    view;
    INT32       num_vertices;

    view.x = view.y = 0;
    view.z = 0;

    /*
     * How Frustum Planes are Calculated:
     * ----------------------------------
     *       __--
     * eye --
     *       --
     *         --
     *
     * 1. Calculate a unit vector parallel to the plane using sin() and cos().
     * 2. Rotate this inwards 90 degrees to create a normal.
     * 3. Calculate the distance.
     * 4. Use this plane to clip :)
     *
     * NOTE: Clipping in world space is the same except that the unit normals
     * must be rotated so that the frustum is pointing where the camera is.
     * Some code like this will be required:
     *
     *      Vec3D_Rotate(&(frustum[1].unit_normal), 0, v_a, 0);
     */

    frustum[0].unit_normal.x = -sin(DEG2RAD(FOV / 2.0));
    frustum[0].unit_normal.y = 0;
    frustum[0].unit_normal.z = cos(DEG2RAD(FOV / 2.0));
    Vec3D_Rotate(&(frustum[0].unit_normal), 0, 90, 0);
    frustum[0].dist = Vec3D_RawDotProduct(&view, &(frustum[0].unit_normal));

    frustum[1].unit_normal.x = sin(DEG2RAD(FOV / 2.0));
    frustum[1].unit_normal.y = 0;
    frustum[1].unit_normal.z = cos(DEG2RAD(FOV / 2.0));
    Vec3D_Rotate(&(frustum[1].unit_normal), 0, -90, 0);
    frustum[1].dist = Vec3D_RawDotProduct(&view, &(frustum[1].unit_normal));

    frustum[2].unit_normal.x = 0;
    frustum[2].unit_normal.y = -sin(DEG2RAD(fov_y / 2.0));
    frustum[2].unit_normal.z = cos(DEG2RAD(fov_y / 2.0));
    Vec3D_Rotate(&(frustum[2].unit_normal), -90, 0, 0);
    frustum[2].dist = Vec3D_RawDotProduct(&view, &(frustum[2].unit_normal));

    frustum[3].unit_normal.x = 0;
    frustum[3].unit_normal.y = sin(DEG2RAD(fov_y / 2.0));
    frustum[3].unit_normal.z = cos(DEG2RAD(fov_y / 2.0));
    Vec3D_Rotate(&(frustum[3].unit_normal), 90, 0, 0);
    frustum[3].dist = Vec3D_RawDotProduct(&view, &(frustum[3].unit_normal));

    num_vertices = ClipPolyToPlane(v_2, v, &frustum[0], 4);
    num_vertices = ClipPolyToPlane(v, v_2, &frustum[1], num_vertices);
    num_vertices = ClipPolyToPlane(v_2, v, &frustum[2], num_vertices);
    num_vertices = ClipPolyToPlane(v, v_2, &frustum[3], num_vertices);

    return num_vertices;
}


/*****************************************************************************
 World Rendering
*****************************************************************************/

static void PerspectiveProject(VERTEX2D *out, VERTEX3D *in, INT32 num_vertices, float xo, float yo, float zo)
{
    INT32  i, xs, ys;
    float   x, y;

    xs = xres / 2;
    ys = yres / 2;

    for (i = 0; i < num_vertices; i++)
    {
        x = (in[i].x + xo) * dphys / (in[i].z + zo);
        y = (in[i].y + yo) * dphys / (in[i].z + zo);

        out[i].x = x + xs;
        out[i].y = y + ys;
    }
}

static void CameraTransform(VERTEX3D *v, INT32 num_v)
{
    INT32   i;
    float   cx, cy, cz;

    for (i = 0; i < num_v; i++)
    {
        cx = v[i].x - x;
        cy = v[i].y - 0;
        cz = v[i].z - z;

        /*
         * Yaw is the only thing we're concerned with
         */

        v[i].x = cz * sin(DEG2RAD(v_a)) + cx * cos(DEG2RAD(v_a));
        v[i].y = cy;
        v[i].z = cz * cos(DEG2RAD(v_a)) - cx * sin(DEG2RAD(v_a));
    }
}

static void DrawWall(BSP *wall)
{
    VERTEX3D    vlist[4*4];
    VERTEX2D    olist[4*4];
    INT32       i;

    vlist[0].x = wall->x1;
    vlist[0].y = WALLHEIGHT / 2;
    vlist[0].z = wall->y1;
    vlist[1].x = wall->x2;
    vlist[1].y = WALLHEIGHT / 2;
    vlist[1].z = wall->y2;
    vlist[2].x = wall->x2;
    vlist[2].y = -(WALLHEIGHT / 2);
    vlist[2].z = wall->y2;
    vlist[3].x = wall->x1;
    vlist[3].y = -(WALLHEIGHT / 2);
    vlist[3].z = wall->y1;

    CameraTransform(vlist, 4);
    i = Clip(vlist);
    PerspectiveProject(olist, vlist, i, 0, 0, 0);

    if (olist[0].x > olist[1].x)
        return;

    Polygon_DrawSolid(0, 0, wall->color, xres, vidbuf, olist, i);
}

static INT32 WallIsFacingUs(BSP *wall)
{
    INT32   wallnorm_x, wallnorm_y;
    INT32   viewvec_x, viewvec_y;
    INT32   dot;

    wallnorm_x = -(wall->y2 - wall->y1);
    wallnorm_y = (wall->x2 - wall->x1);

    viewvec_x = x - wall->x1;
    viewvec_y = z - wall->y1;

    dot = wallnorm_x * viewvec_x + wallnorm_y * viewvec_y;
        
    if (dot <= 0)
        return 1;
    return 0;
}

static void DoWall(BSP *wall)
{
    if (wall == NULL)
        return;

    /*
     * Traverse the tree properly (back to front)
     */

    if (WallIsFacingUs(wall))
    {
        DoWall(wall->f);    // if wall is facing us, do far walls first
        DrawWall(wall);
        DoWall(wall->n);
    }
    else
    {
        DoWall(wall->n);
        DoWall(wall->f);
    }
}

static void RenderWorld()
{
    DoWall(tree);
}


/*****************************************************************************
 Video Code
*****************************************************************************/

static int GetColor()
{
    color++;
    color &= 0xf;   // first 16 colors
    if (color == 0 || color == 8)   // but not these!
        color++;

    return color;
}

static void Blit()
{
    UINT32  i, j, vd, vs;

    for (i = vd = vs = 0; i < lfb.yres; i++)
    {
        for (j = 0; j < lfb.xres / 4; j ++, vd += 4, vs += 4)
            _farpokel(lfb.selector, vd, *((unsigned int *) (vidbuf + vs)));
        vd += lfb.pitch - lfb.xres;
    }
}

static void TextMode()
{
    __dpmi_regs r;

    r.x.ax = 0x0003;
    __dpmi_int(0x10, &r);
}

static INT32 VideoMode(INT32 xres, INT32 yres)
{
    UINT16  mode;

    if (VESALib_CheckVESA())
    {
        fprintf(stderr, "VESA 2.0 is not available!\n");
        return 1;
    }

    VESALib_CreateModeList();

    if (VESALib_FindMode(xres, yres, 8, &mode))
    {
        fprintf(stderr, "Video mode not supported: %dx%dx8\n", xres, yres);
        return 1;
    }

    if ((vidbuf = calloc(xres * yres, sizeof(UINT8))) == NULL)
    {
        fprintf(stderr, "Out of memory!\n");
        return 1;
    }

    if (VESALib_SetMode(mode))
    {
        fprintf(stderr, "Unable to set video mode: %dx%dx8\n", xres, yres);
        return 1;
    }

    if (VESALib_MapLinearFrameBuffer())
    {
        TextMode();
        fprintf(stderr, "Unable to map linear frame buffer!\n");
        return 1;
    }

    VESALib_GetLinearFrameBufferInfo(&lfb);

    return 0;
}


/*****************************************************************************
 BSP Management
*****************************************************************************/

/*
 * FreeBSPTree(): Frees all memory occupied by the BSP tree
 */

static void FreeBSPTree(BSP *b)
{
    if (b == NULL)
        return;

    FreeBSPTree(b->f);
    FreeBSPTree(b->n);

    free(b);
}

/*
 * KillList(): Recursively deletes a linked list of nodes
 */

static void KillList(NODE *n)
{
    if (n == NULL)
        return;

    KillList(n->next);

    free(n);
}

/*
 * FindNode(): Find a node in the node list based on its ID and return a
 * pointer to it
 */

static NODE *FindNode(int id)
{
    NODE    *n;

    for (n = node_list; n != NULL; n = n->next)
    {
        if (n->id == id)
            break;
    }

    if (n == NULL && id != 0)
        printf("error: node not found (%d)\n", id);
    return n;   // this will return NULL if there was no match
}

/*
 * MakeBSPNode(): Creates a BSP node. Works recursively
 */

static void MakeBSPNode(BSP **b, NODE *n)
{
    if (n == NULL)
    {
        *b = NULL;
        return;
    }

    if ((*b = malloc(sizeof(BSP))) == NULL)
    {
        FreeBSPTree(tree);
        KillList(node_list);
        fprintf(stderr, "r: out of memory building BSP tree\n");
        exit(1);
    }

    (*b)->x1 = n->x1;
    (*b)->x2 = n->x2;
    (*b)->y1 = n->y1;
    (*b)->y2 = n->y2;
    (*b)->color = n->color;

    MakeBSPNode(&((*b)->f), FindNode(n->f_id));
    MakeBSPNode(&((*b)->n), FindNode(n->n_id));
}

/*
 * BuildBSPTree(): Constructs a BSP tree based on the information in the node
 * list
 */

static void BuildBSPTree()
{
    MakeBSPNode(&tree, node_list);
}


/*****************************************************************************
 main()
*****************************************************************************/

int main(int argc, char **argv)
{
    FILE    *fp;
    INT32   quit;
    UINT8   move = 0;   // left | right | up | down
    UINT8   k;

    if (argc <= 1)
    {
        puts("r: no input file, nothing to do");
        exit(0);
    }

    /*
     * Read in all of the nodes and then build a BSP tree
     *
     * NOTE: First node must be the root
     */

    if ((fp = fopen(argv[1], "r")) == NULL)
    {
        fprintf(stderr, "r: unable to open %s\n", argv[1]);
        exit(0);
    }

    color = 1;
    while (1)
    {
        NODE    *n, *nsearch;
        int     id, f_id, n_id, x1, y1, x2, y2;

        fscanf(fp, "%d,%d,%d:%d,%d-%d,%d", &id, &f_id, &n_id, &x1, &y1, &x2, &y2);
        if (x1 == 0 && y1 == 0 && x2 == 0 && y2 == 0)   // end of list!
            break;

        if ((n = malloc(sizeof(NODE))) == NULL)
        {
            fclose(fp);
            KillList(node_list);
            fprintf(stderr, "r: out of memory!\n");
            exit(1);
        }

        n->next = NULL; // link into end of list
        n->id = id;
        n->f_id = f_id;
        n->n_id = n_id;
        n->x1 = x1;
        n->y1 = y1;
        n->x2 = x2;
        n->y2 = y2;

        n->color = GetColor(); //color++;

        if (node_list == NULL)  // first node in list?
            node_list = n;      // attach directly to list
        else
        {
            for (nsearch = node_list; nsearch->next != NULL; nsearch = nsearch->next)
                ;
            nsearch->next = n;
        }
    }
    fclose(fp);

    BuildBSPTree();

    /*
     * Set video mode
     */

    if (argc <= 3)
    {
        xres = 640;
        yres = 480;
    }
    else
    {
        xres = atoi(argv[2]);
        yres = atoi(argv[3]);
    }

    dphys = (xres / 2) / tan(DEG2RAD(FOV / 2));
    fov_y = atan((yres / 2) / dphys) * 2;
    fov_y = RAD2DEG(fov_y);

    if (KeybInit())
    {
        KillList(node_list);
        FreeBSPTree(tree);
        exit(1);
    }

    if (VideoMode(xres, yres))
    {
        KillList(node_list);
        FreeBSPTree(tree);
        exit(1);
    } 

    /*
     * Let there be BSP!
     */

    quit = 0;
    while (!quit)
    {
        memset(vidbuf, 0, xres * yres); // erase

        k = KeybGetKey();
        switch (k)
        {
        case KEYB_KEYP_P:
            RenderWorld();
            ScreenShot(vidbuf, xres, yres);
            break;
        case KEYB_KEYP_ESC:
            quit = 1;
            break;
        case KEYB_KEYP_UP:
            move |= 2;
            move &= ~1;
            break;
        case KEYB_KEYR_UP:
            move &= ~2;
            break;
        case KEYB_KEYP_DOWN:
            move |= 1;
            move &= ~2;
            break;
        case KEYB_KEYR_DOWN:
            move &= ~1;
            break;
        case KEYB_KEYP_LEFT:
            move |= 8;
            move &= ~4;
            break;
        case KEYB_KEYR_LEFT:
            move &= ~8;
            break;
        case KEYB_KEYP_RIGHT:
            move |= 4;
            move &= ~8;
            break;
        case KEYB_KEYR_RIGHT:
            move &= ~4;
            break;
        default:
            break;
        }

        if (move & 2)   // Up
        {
            x += sin(DEG2RAD(-v_a))*5;
            z += cos(DEG2RAD(v_a))*5;
        }
        if (move & 1)   // Down
        {
            x -= sin(DEG2RAD(-v_a))*5;
            z -= cos(DEG2RAD(v_a))*5;
        }
        if (move & 8)   // Left
            v_a += 1;
        if (move & 4)   // Right
            v_a -= 1;

        RenderWorld();

        while (inportb(0x3da) & 8)      ;
        while (!(inportb(0x3da) & 8))   ;
        Blit();
    }

    /*
     * Quit!
     */

    TextMode();
    KeybDeinit();
    KillList(node_list);
    FreeBSPTree(tree);

    return 0;
}
